/*
 * Copyright (c) 2016,2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-aarch64.c
 * @brief unwinding call stacks, functions specific for aarch64
 */
#include "crash-stack.h"
#include "shared/log.h"
#include "shared/util.h"

#include <elf.h>
#include <sys/user.h>
#include <string.h>
#include <stdio.h>

static struct user_regs_struct g_registers;	///< static storage for registers

/**
 * @brief Important registers for unwinding stack on aarch64
 */
struct Regs {
	Elf64_Addr x29;		///< x29 register contains frame pointer
	Elf64_Addr x30;		///< x30 register contains return address
	Elf64_Addr pc;		///< register: program counter
	Elf64_Addr sp;		///< register: stack pointer
};

#define REG_X29 29		///< number of x29 register
#define REG_X30 30		///< number of x30 register

typedef struct Regs Regs;	///< convenience type definition
static Regs g_regs;			///< static storage for target register values

void *_crash_stack_get_memory_for_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void _crash_stack_set_registers(void *regbuf)
{
  struct user_regs_struct *regs = regbuf;

  memcpy(_get_place_for_register_value("sp", 0), &regs->sp, sizeof(regs->sp));
  memcpy(_get_place_for_register_value("pc", 0), &regs->pc, sizeof(regs->pc));
  memcpy(_get_place_for_register_value("x29", 0), &regs->regs[29], sizeof(regs->regs[29]));
  memcpy(_get_place_for_register_value("x30", 0), &regs->regs[30], sizeof(regs->regs[30]));
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	if (strcmp(regname, "pc") == 0 || REG_AARCH64_PC == regnum)
		return &g_regs.pc;
	else if (strcmp(regname, "sp") == 0 || REG_AARCH64_SP == regnum)
		return &g_regs.sp;
	else if (strcmp(regname, "x29") == 0 || REG_X29 == regnum)
		return &g_regs.x29;
	else if (strcmp(regname, "x30") == 0 || REG_X30 == regnum)
		return &g_regs.x30;
	return NULL;
}

unsigned long _get_register_value(int n)
{
	return g_registers.regs[n];
}

#define PSR_MODE32_BIT  0x00000010
#define PSR_MODE_MASK   0x0000000f

void _crash_stack_print_regs(json_object *jobj)
{
#define ADD_NUM_REG(jobj, index, value) \
	{ \
	    char *s; \
	    if (asprintf(&s, "x%-2zd", index) == -1) { \
	            _E("Cannot print register name: x%-2zd", index); \
	    } else { \
		    char value_str[16+3]; \
		    if (sprintf(value_str, "0x%016llx", value) != -1) \
			    ADD_REG_TO_JSON_ARRAY(jobj, s, value_str); \
	    } \
	    free(s); \
	} while(0)

	json_object *j_regs = json_object_new_array();

	for (size_t i = 0; i <= 30; i++)
		ADD_NUM_REG(j_regs, i, g_registers.regs[i]);

	struct reg {
		const char *name;
		unsigned long long int value;
	} regs[] = {{"xr", g_registers.regs[8]},
	            {"ip0", g_registers.regs[16]},
	            {"ip1", g_registers.regs[17]},
	            {"pr", g_registers.regs[18]},
	            {"fp", g_registers.regs[29]},
	            {"lr", g_registers.regs[30]},
	            {"pc", g_registers.pc},
	            {"sp", g_registers.sp}};

	for (size_t i = 0; i < ARRAY_SIZE(regs); i++) {
		char value_str[16+3];
		if (sprintf(value_str, "0x%016llx", regs[i].value) == -1) {
			_E("Cannot print register value: %m");
			continue;
		}
		ADD_REG_TO_JSON_ARRAY(j_regs, regs[i].name, value_str);
	}

	json_object_object_add(jobj, K_REGISTERS, j_regs);
}
