/*
 * Copyright (c) 2016,2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-x86_64.c
 * @brief unwinding call stacks, functions specific for x86_64
 */
#include "crash-stack.h"
#include "shared/log.h"
#include "shared/util.h"

#include <sys/user.h>
#include <stdio.h>
#include <string.h>

struct user_regs_struct g_registers;	///< static storage for registers

void *_crash_stack_get_memory_for_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	/* Function is not used for x86 anymore, make it dummy
	 * It is still called by generic __get_registers_core function
	 * The return value is checked for NULL in __get_registers_core.
	 */
	return NULL;
}

void _crash_stack_set_registers(void *regbuf)
{
	/* Make it dummy as it not used for x86 anymore
	 * However it is still called by generic __get_registers functions
	 */
}

unsigned long _get_register_value(int n)
{
	const static int reg_map[] = {
#if defined(__x86_64__)
		[UNW_X86_64_RAX] = offsetof(struct user_regs_struct, rax) / sizeof(long),
		[UNW_X86_64_RDX] = offsetof(struct user_regs_struct, rdx) / sizeof(long),
		[UNW_X86_64_RCX] = offsetof(struct user_regs_struct, rcx) / sizeof(long),
		[UNW_X86_64_RBX] = offsetof(struct user_regs_struct, rbx) / sizeof(long),
		[UNW_X86_64_RSI] = offsetof(struct user_regs_struct, rsi) / sizeof(long),
		[UNW_X86_64_RDI] = offsetof(struct user_regs_struct, rdi) / sizeof(long),
		[UNW_X86_64_RBP] = offsetof(struct user_regs_struct, rbp) / sizeof(long),
		[UNW_X86_64_RSP] = offsetof(struct user_regs_struct, rsp) / sizeof(long),
		[UNW_X86_64_RIP] = offsetof(struct user_regs_struct, rip) / sizeof(long),
#else
		[UNW_X86_EAX] = offsetof(struct user_regs_struct, eax) / sizeof(long),
		[UNW_X86_EDX] = offsetof(struct user_regs_struct, edx) / sizeof(long),
		[UNW_X86_ECX] = offsetof(struct user_regs_struct, ecx) / sizeof(long),
		[UNW_X86_EBX] = offsetof(struct user_regs_struct, ebx) / sizeof(long),
		[UNW_X86_ESI] = offsetof(struct user_regs_struct, esi) / sizeof(long),
		[UNW_X86_EDI] = offsetof(struct user_regs_struct, edi) / sizeof(long),
		[UNW_X86_EBP] = offsetof(struct user_regs_struct, ebp) / sizeof(long),
		[UNW_X86_ESP] = offsetof(struct user_regs_struct, esp) / sizeof(long),
		[UNW_X86_EIP] = offsetof(struct user_regs_struct, eip) / sizeof(long),
		[UNW_X86_EFLAGS] = offsetof(struct user_regs_struct, eflags) / sizeof(long),
		[UNW_X86_TRAPNO] = offsetof(struct user_regs_struct, orig_eax) / sizeof(long),
#endif
	};
	n = reg_map[n];
	return ((elf_greg_t*)&g_registers)[n];
}

void _crash_stack_print_regs(json_object *jobj)
{
#if defined(__x86_64__)
	struct reg {
		const char *name;
		unsigned long long int value;
	} regs[] = {{"gs", g_registers.gs},
	            {"fs", g_registers.fs},
	            {"es", g_registers.es},
	            {"ds", g_registers.ds},
	            {"rdi", g_registers.rdi},
	            {"rsi", g_registers.rsi},
	            {"rbp", g_registers.rbp},
	            {"rsp", g_registers.rsp},
	            {"rax", g_registers.rax},
	            {"rbx", g_registers.rbx},
	            {"rcx", g_registers.rcx},
	            {"rdx", g_registers.rdx},
	            {"r8", g_registers.r8},
	            {"r9", g_registers.r9},
	            {"r10", g_registers.r10},
	            {"r11", g_registers.r11},
	            {"r12", g_registers.r12},
	            {"r13", g_registers.r13},
	            {"r14", g_registers.r14},
	            {"r15", g_registers.r15},
	            {"rip", g_registers.rip}};

#define ADDR_FORMAT "0x%016llx"
#else
	struct reg {
		const char *name;
		unsigned long int value;
	} regs[] = {{"gs", g_registers.xgs},
	            {"fs", g_registers.xfs},
	            {"es", g_registers.xes},
	            {"ds", g_registers.xds},
	            {"edi", g_registers.edi},
	            {"esi", g_registers.esi},
	            {"ebp", g_registers.ebp},
	            {"esp", g_registers.esp},
	            {"eax", g_registers.eax},
	            {"ebx", g_registers.ebx},
	            {"ecx", g_registers.ecx},
	            {"edx", g_registers.edx},
	            {"eip", g_registers.eip}};

#define ADDR_FORMAT "0x%08lx"
#endif
	json_object *j_regs = json_object_new_array();

	for (size_t i = 0; i < ARRAY_SIZE(regs); i++) {
		char value_str[16+3];
		if (sprintf(value_str, ADDR_FORMAT, regs[i].value) == -1) {
			_E("Cannot print register value: %m");
			continue;
		}
		ADD_REG_TO_JSON_ARRAY(j_regs, regs[i].name, value_str);
	}

	json_object_object_add(jobj, K_REGISTERS, j_regs);
#undef ADDR_FORMAT
}
